"""
Tools specific to computing arc angles
"""

import numpy as np
from scipy.spatial.transform import Rotation

from .rotations import make_homogeneous_transform


def calculate_arc_angles(target_pt, extra_pt, ap_offset=14, degrees=True):
    """
    Compute the arc angles needed for a probe trajectory that intersects 2
    points.

    Note that order matters on the points;
    currently "target" is the intended deep point and "extra" is a point
    at/above the surface.

    This should probably have some coordinate system awareness, and this
    documentation should be expanded to show logic.

    # Returns AP, ML angle
    """
    this_vector = (extra_pt - target_pt) / np.linalg.norm(extra_pt - target_pt)
    phi = np.arcsin(this_vector[0])
    theta = np.arcsin(-this_vector[1] / np.cos(phi))
    return np.rad2deg(theta) + ap_offset, -np.rad2deg(phi)


def transform_matrix_from_angles_and_target(AP, ML, target, rotation=0):
    euler_angles = np.array([np.deg2rad(x) for x in [AP, ML, rotation]])
    R = Rotation.from_euler("XYZ", euler_angles).as_matrix().squeeze()
    return make_homogeneous_transform(R, target)



"""
Functions for saving and loading transforms using SimpleITK.
"""

import numpy as np
import SimpleITK as sitk

from .. import rotations as rot


def save_sitk_transform(
    filename,
    rotation_matrix,
    translation=None,
    transpose_matrix=False,
    legacy=False,
):
    """
    Save transform to a sitk file (readable by Slicer);

    Parameters
    ----------
    filename : string
        filename to save to.
    T : np.array(4,3), np.array(4x4), or np.array(6,)
        Rigid transform t
        o save.
        if a np.array(6,) is passed, a rigid transform is created using
        aind_mri_utils.optimization.crate_rigid_transform.
    invert = bool, optional
        If true, invert the transform before saving.
        Default is False.


    """
    if len(rotation_matrix) == 6:
        R = rot.combine_angles(*rotation_matrix[:3])
        found_translation = rotation_matrix[3:]
    elif rotation_matrix.shape == (4, 4):
        if legacy:
            found_translation = rotation_matrix[3, :3]
        else:
            found_translation = rotation_matrix[:3, 3]
        R = rotation_matrix[:3, :3]
    elif rotation_matrix.shape == (3, 4) and not legacy:
        R = rotation_matrix[:, :3]
        found_translation = rotation_matrix[:, 3]
    elif rotation_matrix.shape == (4, 3) and legacy:
        R = rotation_matrix[:3, :]
        found_translation = rotation_matrix[3, :]
    elif rotation_matrix.shape == (3, 3):
        R = rotation_matrix
        found_translation = np.zeros(3)
    else:
        raise ValueError("Invalid transform shape and legacy flag")
    if translation is not None:
        found_translation = translation
    if transpose_matrix:
        if not legacy:
            raise ValueError(
                "transpose_matrix only valid for legacy transforms"
            )
        R = R.T
    A = rot.rotation_matrix_to_sitk(R, translation=found_translation)
    sitk.WriteTransform(A, filename)


def load_sitk_transform(
    filename, homogeneous=False, legacy=False, invert=False
):
    """
    Convert a sitk transform file to a 4x3 numpy array.

    Parameters
    ----------
    filename : string
        filename to load from.
    homogeneous : bool, optional
        If True, return a 4x4 homogeneous transform matrix. Default is False.
    legacy : bool, optional
        If True, return a 4x3 transform matrix with the translation as the
        last row. Default is False

    Returns
    -------
    R: np.array(N,M)
        Rotation matrix. For three dimensional transforms: np.array(3,3). If
        homogeneous: np.array(4, 4), if legacy: np.array(4, 3)
    translation: np.array(L,)
        Translation vector. Not returned if legacy is True.
    center: np.array(L,)
        Center of rotation. Not returned if legacy is True.
    """
    A = sitk.ReadTransform(filename)
    if invert:
        A = A.GetInverse()
    R, translation, center = rot.sitk_to_rotation_matrix(A)
    if legacy:
        R = np.vstack((R, translation))
        return R
    if homogeneous:
        if not np.allclose(center, 0):
            raise NotImplementedError(
                "homogeneous only valid for transforms with center at 0"
            )
        if legacy:
            raise ValueError(
                "homogeneous only valid for non-legacy transforms"
            )
        R = rot.make_homogeneous_transform(R, translation)
    return R, translation, center



# -*- coding: utf-8 -*-
"""
Created on Tue Feb  7 15:30:09 2023

@author: yoni.browning
"""

import unittest

import numpy as np

from aind_mri_utils import rotations


class RotationsTest(unittest.TestCase):
    def test_define_euler_rotation(self) -> None:
        R = rotations.define_euler_rotation(0, 0, 0, degrees=True)
        self.assertTrue(np.array_equal(R.as_matrix(), np.eye(3)))

    def test_rotate_about(self) -> None:
        # Test 1: no rotation
        pt = np.array([1, 2, 3])
        pivot = np.array([2, 3, 4])
        R = rotations.define_euler_rotation(0, 0, 0, degrees=True)
        X = rotations.rotate_about(pt, R, pivot)
        self.assertTrue(np.array_equal(X, pt))
        # Test 2: 360 rotation
        pt = np.array([1, 2, 3])
        pivot = np.array([2, 3, 4])
        R = rotations.define_euler_rotation(360, 360, 360, degrees=True)
        X = rotations.rotate_about(pt, R, pivot)
        self.assertTrue(np.array_equal(X, pt))
        # Test 3: Numerical Error
        pt = np.array([1, 2, 3])
        pivot = np.array([2, 3, 4])
        R = rotations.define_euler_rotation(
            np.pi * 2, np.pi * 2, np.pi * 2, degrees=False
        )
        X = rotations.rotate_about(pt, R, pivot)
        self.assertFalse(np.array_equal(X, pt))
        self.assertTrue(np.all(X - pt < 0.02))
        # Test4: with translation
        pt = np.array([1, 2, 3])
        pivot = np.array([2, 3, 4])
        R = rotations.define_euler_rotation(360, 360, 360, degrees=True)
        translate = np.array((1, 1, 1))
        X = rotations.rotate_about_and_translate(
            pt, R, pivot, np.array(translate)
        )
        self.assertTrue(np.array_equal(X, pt + translate))
        # Test 5: More than one point
        pt = np.array([[1, 2, 3], [1, 2, 3]])
        pivot = np.array([2, 3, 4])
        R = rotations.define_euler_rotation(0, 0, 0, degrees=True)
        X = rotations.rotate_about(pt, R, pivot)
        self.assertTrue(X.shape[1] == pt.shape[1])
        self.assertTrue(np.array_equal(X[0, :], X[1, :]))
        # Test 6: non-identity
        pt = np.array([0, 1, 0], dtype=np.float64)
        R = rotations.define_euler_rotation(90, 0, 0, degrees=True)
        pivot = np.array([0, 0, 0])
        X = rotations.rotate_about(pt, R, pivot)
        self.assertTrue(np.all(X - np.array([0.0, 0.0, 1.0]) < 0.00001))

    def test_scipy_rotation_to_sitk(self) -> None:
        R = rotations.define_euler_rotation(90, 0, 0)
        center = np.array((-1, 0, 0))
        translation = np.array((1, 0, 0))
        trans = rotations.scipy_rotation_to_sitk(
            R, center=center, translation=translation
        )
        self.assertTrue(np.array_equal(trans.GetTranslation(), translation))
        self.assertTrue(np.array_equal(trans.GetFixedParameters(), center))
        self.assertTrue(
            np.array_equal(
                R.as_matrix().reshape((9,)),
                np.array(trans.GetParameters()[:9]),
            )
        )

    def test_rotation_matrix_from_vectors(self) -> None:
        a = np.array([1, 0, 0])
        b = np.array([1, 1, 0]) / np.sqrt(2)
        rotmat = rotations.rotation_matrix_from_vectors(a, b)
        self.assertTrue(np.allclose(rotmat @ a, b))
        rotmat = rotations.rotation_matrix_from_vectors(a, a)
        self.assertTrue(np.allclose(rotmat, np.eye(a.size)))
        rotmat = rotations.rotation_matrix_from_vectors(a, -a)
        self.assertTrue(np.allclose(rotmat, -np.eye(a.size)))
        self.assertRaises(
            ValueError, rotations.rotation_matrix_from_vectors, a, np.zeros(1)
        )


if __name__ == "__main__":
    unittest.main()
